"""
Copied from: https://github.com/python/typeshed/blob/master/stdlib/pkgutil.pyi
Copyright: https://github.com/python/typeshed/blob/master/LICENSE

"""
from typing import Any

def resolve_name(name: str) -> Any: ...
